package at.jku.students;

import at.jku.ssw.ListNode;
import at.jku.ssw.DotMaker;
import at.jku.ssw.Iterator;
import at.jku.ssw.List;
import java.lang.reflect.Array;

/**
 * A linked list of objects.
 */
public class LinkedList extends List {

    private ListNode head = null;
    private ListNode tail = null;
    private int size = 0;

    /**
     * Returns the value at position index.
     */
    public Object get(int index) {
        if (index < 0) {
            throw new IndexOutOfBoundsException();
        }
        ListNode p = head;
        int pos = 0;
        while (p != null && pos < index) {
            pos++;
            p = p.next;
        }
        if (p != null) {
            return p.value;
        } else {
            throw new IndexOutOfBoundsException();
        }
    }

    /**
     * Returns the index of the first occurrence of the value.
     */
    public int indexOf(Object value) {
        int index = 0;
        ListNode p = head;
        while (p != null) {
            if (p.value.equals(value)) {
                return index;
            }
            index++;
            p = p.next;
        }
        return -1;
    }

    /**
     * Returns the number of values in this list.
     */
    public int size() {
        return size;
    }

    /**
     * Removes the value at position index.
     */
    public Object remove(int index) {
        if (index < 0 || index >= size) {
            throw new IndexOutOfBoundsException();
        }
        if (index == size - 1) {
            return removeLast();
        }

        int at = 0;
        ListNode p = head;
        while (at < index) {
            at++;
            p = p.next;
        }
        p.next.prev = p.prev;
        if (head == p) {
            head = p.next;
        } else {
            p.prev.next = p.next;
        }
        size--;
        return p.value;
    }

    /**
     * Returns an iterator of the values in this list.
     */
    public Iterator iterator() {
        LinkedList copy = new LinkedList();
        ListNode p = head;
        while (p != null) {
            copy.insertLast(p.value);
            p = p.next;
        }
        return new LinkedListIterator(copy.head);
    }

    /**
     * Returns an array containing all of the elements in this list in proper
     * sequence (from first to last element); the runtime type of the returned
     * array is that of the specified array. If the list fits in the specified
     * array, it is returned therein. Otherwise, a new array is allocated with
     * the runtime type of the specified array and the size of this list.
     */
    public Object[] toArray(Object[] array) {
        int size = size();
        Object[] result = array;
        if (size > result.length) {
            result = (Object[]) Array.newInstance(array.getClass().getComponentType(), size);
        }
        Iterator it = iterator();
        int index = 0;
        while (it.hasNext()) {
            result[index++] = it.next();
        }
        return result;
    }

    @Override
    public void insert(int index, Object value) {
        if (index < 0 || index > size()) {
            throw new IndexOutOfBoundsException();
        }

        // fast insert at the end of the list
        if (index == size) {
            insertLast(value);
            return;
        }

        assert tail != null;
        assert head != null;

        int idx = 0;
        ListNode p = head;
        ListNode n = new ListNode(value);

        while (p != null && idx < index) {
            idx++;
            p = p.next;
        }

        assert p != null;

        if (p.prev != null) {
            p.prev.next = n;
        } else {
            head = n;
        }
        n.prev = p.prev;
        n.next = p;
        p.prev = n;
        
        size++;
    }

    @Override
    public void insertLast(Object value) {
        ListNode n = new ListNode(value);
        if (tail != null) {
            tail.next = n;
        } else {
            head = n;
        }
        n.prev = tail;
        tail = n;
        size++;
    }

    @Override
    public Object removeLast() {
        if (tail == null) {
            return null;
        }
        Object value = tail.value;

        tail = tail.prev;
        if (tail != null) {
            tail.next = null;
        } else {
            head = null;
        }
        size--;
        return value;
    }

    @Override
    public int lastIndexOf(Object value) {
        int idx = size() - 1;
        ListNode p = tail;
        while (p != null) {
            // value found
            if (p.value.equals(value)) {
                return idx;
            }

            // move to previous node
            idx--;
            p = p.prev;
        }

        // not found
        return -1;
    }
    
    public String makeDot() {
        return DotMaker.makeDotForLinkedList(head);
    }
}
